#!/usr/bin/perl -w

#
# dbmapreduce.pm
# Copyright (C) 1991-2008 by John Heidemann <johnh@isi.edu>
# $Id$
#
# This program is distributed under terms of the GNU general
# public license, version 2.  See the file COPYING
# in $dblibdir for details.
#


package Jdb::Filter::dbmapreduce;

=head1 NAME

dbmapreduce - reduce all input rows with the same key

=head1 SYNOPSIS

usage: $0 [-dMS] [-k KeyField] [-C code] [ReduceCommand [ReduceArguments...]]

=head1 DESCRIPTION

Group input data by KeyField,
then reduce each group.
The reduce function can be an external program
given by ReduceCommand and ReduceArguments,
or an Perl subroutine given by CODE.

This program thus implements Google-style map/reduce,
but executed sequentially.
With the C<-M> option, the reducer is given multiple groups
(as with Google), but not in any guaranteed order
(while Google guarantees they arive in lexically sorted order).
However without it, a stronger invarient
is provided: each reducer is given exactly one group.

By default the KeyField is the first field in the row.
Unlike Hadoop streaming, the -k KeyField option can explicitly
name where the key is in any column of each input row.

The KeyField used for each Reduce is added to the beginning of each 
row of reduce output.

The reduce function can do anything, emitting one or more output records.
However, it must have the same output field separator as the input data.
(In some cases this case may not occur, for example, input data with -FS
and a reduce function of dbstats.  This bug needs to be fixed in the future.)


Assumptions and requirements:	

By default, data can be provided in arbitrary order
and the program consumes O(number of unique tags) memory,
and O(size of data) disk space.

With the -S option, data must arrive group by tags (not necessarily sorted),
and the program consumes O(number of tags) memory and no disk space.
The program will check and abort if this precondition is not met.

With two -S's, program consumes O(1) memory, but doesn't verify
that the data-arrival precondition is met.

=head1 OPTIONS

=over

=item B<-k> or B<--key> KeyField

specify which column is the key for grouping (default: the first column)

=item B<-S> or B<--pre-sorted>

Assume data is already grouped by tag.
Provided twice, it removes the validiation of this assertion.

=item B<-M> or B<--multiple-ok>

Assume the ReduceCommand can handle multiple grouped keys,
and the ReduceCommand is responsible for outputting the  
with each output row.
(By default, a separate ReduceCommand is run for each key,
and dbmapreduce adds the key to each output row.)

=item B<-C> or B<--code>

Provide Perl code that implements the reduce function.
The provided code should be an anonymous sub
that creates a Jdb Filter that implements the reduce object.

=back

=for comment
begin_standard_jdb_options

This module also supports the standard jdb options:

=over

=item B<-d>

Enable debugging output.

=item B<-i> or B<--input> InputSource

Read from InputSource, typically a file name, or C<-> for standard input,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<-o> or B<--output> OutputDestination

Write to OutputDestination, typically a file name, or C<-> for standard output,
or (if in Perl) a IO::Handle, Jdb::IO or Jdb::BoundedQueue objects.

=item B<--autorun> or B<--noautorun>

By default, programs process automatically,
but Jdb::Filter objects in Perl do not run until you invoke
the run() method.
The C<--(no)autorun> option controls that behavior within Perl.

=item B<--help>

Show help.

=item B<--man>

Show full manual.

=back

=for comment
end_standard_jdb_options


=head1 SAMPLE USAGE

=head2 Input:

    #h experiment duration
    ufs_mab_sys 37.2
    ufs_mab_sys 37.3
    ufs_rcp_real 264.5
    ufs_rcp_real 277.9

=head2 Command:

    cat DATA/stats.jdb | dbmapreduce -k experiment dbstats duration

=head2 Output:

    #h      experiment      mean    stddev  pct_rsd conf_range      conf_low       conf_high        conf_pct        sum     sum_squared     min     max     n
    ufs_mab_sys     37.25 0.070711 0.18983 0.6353 36.615 37.885 0.95 74.5 2775.1 37.2 37.3 2
    ufs_rcp_real    271.2 9.4752 3.4938 85.13 186.07 356.33 0.95 542.4 1.4719e+05 264.5 277.9 2
    #  | dbmapreduce -k experiment dbstats duration


=head1 SEE ALSO

L<Jdb>.
L<dbmultistats>
L<dbrowsplituniq>


=head1 CLASS FUNCTIONS

=cut

@ISA = qw(Jdb::Filter);
$VERSION = 2.0;

use strict;
use Pod::Usage;
use threads;
use threads::shared;
use Carp;

use Jdb::Filter;
use Jdb::IO::Reader;
use Jdb::IO::Writer;
use Jdb::Filter::dbsubprocess;
use Jdb::Filter::dbpipeline qw(dbpipeline_filter dbpipeline_open2 dbsort dbsubprocess);


=head2 new

    $filter = new Jdb::Filter::dbmapreduce(@arguments);

Create a new dbmapreduce object, taking command-line arguments.

=cut

sub new ($@) {
    my $class = shift @_;
    my $self = $class->SUPER::new(@_);
    bless $self, $class;
    $self->set_defaults;
    $self->parse_options(@_);
    $self->SUPER::post_new();
    return $self;
}


=head2 set_defaults

    $filter->set_defaults();

Internal: set up defaults.

=cut

sub set_defaults ($) {
    my($self) = @_;
    $self->SUPER::set_defaults();
    $self->{_key_column} = undef;
    $self->{_pre_sorted} = 0;
    $self->{_reduce_generator_code} = undef;
    $self->{_reduce_generator} = undef;
    $self->{_map_aware_reduce} = undef;
    $self->{_external_command_argv} = [];
    $self->{_reducer_generator_code} = undef;
}

=head2 parse_options

    $filter->parse_options(@ARGV);

Internal: parse command-line arguments.

=cut

sub parse_options ($@) {
    my $self = shift @_;

    my(@argv) = @_;
    $self->get_options(
	\@argv,
 	'help|?' => sub { pod2usage(1); },
	'man' => sub { pod2usage(-verbose => 2); },
	'autorun!' => \$self->{_autorun},
	'close!' => \$self->{_close},
	'C|code=s' => \$self->{_reducer_generator_code},
	'd|debug+' => \$self->{_debug},
	'i|input=s' => sub { $self->parse_io_option('input', @_); },
	'k|key=s' => \$self->{_key_column},
	'log!' => \$self->{_logprog},
	'M|multiple-ok!' => \$self->{_map_reduce_aware},
	'o|output=s' => sub { $self->parse_io_option('output', @_); },
	'S|pre-sported+' => \$self->{_pre_sorted},
	'saveoutput=s' => \$self->{_save_output},
	) or pod2usage(2);
    push (@{$self->{_external_command_argv}}, @argv);
}

=head2 setup

    $filter->setup();

Internal: setup, parse headers.

=cut

sub setup ($) {
    my($self) = @_;

    #
    # what are we running?
    #
    # Figure it out, and generate a 
    # $self->{_reducer_generator_sub} that creates a
    # filter object that will be passed to dbpipeline_open2
    # to reduce one (or many, if map_reduce_aware) keys.
    #
    if ($#{$self->{_external_command_argv}} >= 0) {
	# external command
	my @argv = @{$self->{_external_command_argv}};
	my $reducer_sub = sub {
	    dbsubprocess('--nolog', @argv);
	};
	$self->{_reducer_generator_sub} = $reducer_sub;
    } elsif (defined($self->{_reducer_generator_code})) { 
        my $sub_code;
	$sub_code = 
	    "use Jdb::Filter::dbpipeline qw(:all);\n" . 
	    '$reducer_sub = sub {' . "\n" .
	    $self->{_reducer_generator_code} .
	    "\n\t;\n};\n";
	my $reducer_sub;
	print "$sub_code" if ($self->{_debug});
	eval $sub_code;
	$@ && croak $self->{_prog} . ": error evaluating user-provided reducer sub:\n$sub_code\nerror is: $@.\n";
	$self->{_reducer_generator_sub} = $reducer_sub;
    } else {
	croak $self->{_prog} . ": reducer not specified.\n";
    };

    $self->finish_io_option('input', -comment_handler => $self->create_pass_comments_sub);
    if (defined($self->{_key_column})) {
	$self->{_key_coli} = $self->{_in}->col_to_i($self->{_key_column});
	croak $self->{_prog} . ": key column " . $self->{_key_column} . " is not in input stream.\n"
	    if (!defined($self->{_key_coli}));
    } else {
	# default to first column
	$self->{_key_coli} = 0;
	$self->{_key_column} = $self->{_in}->i_to_col(0);
    };

    # do we need to group the keys for the user?
    if (!$self->{_pre_sorted}) {
	# just do lexical sort
        my($new_reader, $new_thread) = dbpipeline_filter($self->{_in}, dbsort($self->{_key_column}));
	$self->{_pre_sorted_in} = $self->{_in};
	$self->{_in} = $new_reader;
	$self->{_in_thread} = $new_thread;
	#
	# We detach the sorter here.
	#
	# This is OK, sort of, but seems broken.
	# We do it because otherwise we block joining the thread
	# at the end of $self->run.
	# We shouldn't, because the sorter should always terminate.
	# And we KNOW it's run it's finish(), because it must have
	# generated EOF.
	#
	# So this may indicate a bug in sort or dbpipline_filter
	# with a lingering thread.
	# But to debug later, as detaching seems like a plausible work-around
	# and not TOO evil.
	#
	$new_thread->detach;
    };

    #
    # setup the postprocessing thread
    #
    $self->_setup_postprocessor;

    $self->{_reducer_invocation_count} = 0;
    $SIG{'PIPE'} = sub {
        die $self->{_prog} . ": cannot run external dbmapreduce reduce program (" . join(" ", @{$self->{_external_command_argv}}) . ")\n";
    };

}

=head2 _setup_postprocessor

    _setup_postprocessor

(internal)  One thread that merges the output.
C<_postprocessor_queue> should be handed queues to jdb reader objects
from each reducer.
We setup the output, suppress all but the first header,
and add in the keys if necessary.

=cut

sub _setup_postprocessor {
    my($self) = @_;

    $self->{_postprocessor_queue} = new Jdb::BoundedQueue;
    my $postprocessor_thread = threads->new( sub {
	    $self->_reduce_postprocessor;
	} );
    $self->{_reducer_postprocess_thread} = $postprocessor_thread;
}

=head2 _reduce_postprocessor 

    _reduce_postprocessor 

Parse reduce output, adding our tag and stripping comments
when necessary.
We return to the caller, assuming they will exit, if necessary.

=cut

sub _reduce_postprocessor {
    my($self) = @_;

    my $queue_of_queues = $self->{_postprocessor_queue};
    my $queue_from_reducer;
    my $writer_setup = undef;
    my $map_aware_reduce = $self->{_map_aware_reduce};
    my $write_fastpath_sub = undef;

    my $reducer_info;
    while ($reducer_info = $queue_of_queues->dequeue) {
	last if (!defined($reducer_info));

	my($from_reduce_queue, $reducer_thread, $current_key) = @$reducer_info;

	#
	# convert the reducer's queue to a reader object
	#
	my $from_reducer = new Jdb::IO::Reader(-queue => $from_reduce_queue,
	    -comment_handler => $self->create_pass_comments_sub);

	# setup the writer, if necessary
	if (!$writer_setup) {
	    # maybe need to add in the key
	    my @cols = @{$from_reducer->cols};
	    if (!$map_aware_reduce) {
		# yes, push it on the front
		unshift(@cols, $self->{_key_column});
	    };
	    $self->finish_io_option('output', -clone => $from_reducer, -cols => \@cols);
	    $write_fastpath_sub = $self->{_out}->fastpath_sub();
	    $writer_setup = 1;
	    # Save the first reader object forever
	    # so we can check for compatibility.
	    $self->{_first_reducer_reader} = $from_reducer;
	} else {
	    # Writer already existed, BUT out of paranoia we check
	    # that other reducers don't change the schema.
	    croak $self->{_prog} . ": reducers have non-identical output schema.\n"
		if ($self->{_first_reducer_reader}->compare($from_reducer) ne 'identical');
	};

	# first, log something if we're debugging
	$self->{_out}->write_comment("mapredcue postprocessor: new reducer instance on key: " . $current_key)
	    if ($self->{_debug});

	# Ok, we have our reader and writer.
	# Copy the output, adding our field if necessary.
	my $fref;
	my $read_fastpath_sub = $from_reducer->fastpath_sub();
	if ($map_aware_reduce) {
	    while ($fref = &$read_fastpath_sub()) {
		&$write_fastpath_sub($fref);
	    };
	} else {
	    # Add our new field for less aware reducers.
	    while ($fref = &$read_fastpath_sub()) {
		unshift (@$fref, $current_key);
		&$write_fastpath_sub($fref);
	    };
	    $self->{_out}->write_comment("mapredcue postprocessor: done with reducer instance on key: " . $current_key)
	    if ($self->{_debug});
	    # now reap the reducer
	    $reducer_thread->join;
	    $from_reduce_queue = undef;
	    $from_reducer = undef;
	};
    };

    # Done with all reducers.
    # Finish up the IO.
    $self->SUPER::finish();
}


=head2 _open_new_key

    _open_new_key

(internal)

=cut

sub _open_new_key {
    my($self, $new_key) = @_;
    $self->{_current_key} = $new_key;
    # If already running and can handle multiple tags, just keep going.
    return if ($self->{_map_aware_reduce} &&
	    defined($self->{_reducer_thread}));

    #
    # make the reducer
    #

    # first, start the reducer
    my $users_reducer = &{$self->{_reducer_generator_sub}}($new_key);
    my($from_reducer_queue, $to_reducer_writer, $reducer_thread) = 
	    dbpipeline_open2([-clone => $self->{_in}], $users_reducer);
    $self->{_current_reducer} = $to_reducer_writer;
    $self->{_current_reducer_fastpath_sub} = $to_reducer_writer->fastpath_sub();

    # now send it to our postprocessor
    share($from_reducer_queue);
    share($reducer_thread);
    my @postprocessor_info = ($from_reducer_queue, $reducer_thread, $new_key);
    $self->{_postprocessor_queue}->enqueue(\@postprocessor_info); 

}

=head2 _close_old_key

    _close_old_key

Internal, finish a tag.

=cut

sub _close_old_key {
    my($self, $key, $final) = @_;
    return if (!defined($key));   # xxx: why do this?
    return if ($self->{_map_aware_reduce} && !$final);  # can keep handling them in the reducer

    croak $self->{_prog} . ": internal error: current key doesn't equal prior key " . $self->{_current_key} . " != key $key \n"
	if ($self->{_current_key} ne $key);
    # finish the reducer
    $self->{_current_reducer}->close;
    # we expect the post-processor to join on the thread to reap it

    # finish the post-processor if we're done
    $self->{_postprocessor_queue}->enqueue(undef)
	if ($final);
}

=head2 run

    $filter->run();

Internal: run over each rows.

=cut
sub run ($) {
    my($self) = @_;

    my $read_fastpath_sub = $self->{_in}->fastpath_sub();
    my $reducer_fastpath_sub;

    # read data
    my($last_key) = undef;
    my $fref;
    my $key_coli = $self->{_key_coli};
    $self->{_key_counts} = {};
    while ($fref = &$read_fastpath_sub()) {
	# print STDERR "data line: " . join("  ", @$fref) . "\n";
        my($key) = $fref->[$key_coli];
    
        if (!defined($last_key) || $key ne $last_key) {
            # start a new one
            # check for out-of-order duplicates
            if ($self->{_pre_sorted} == 1) {
                croak $self->{_prog} . ": single key ``$key'' split into multiple groups, selection of -S was invalid\n"
                    if (defined($self->{_key_counts}{$key}));
                $self->{_key_counts}{$key} = 1;
            };
            # finish off old one?
            if (defined($last_key)) {
                $self->_close_old_key($last_key);
            };
            $self->_open_new_key($key);
            $last_key = $key;
	    $reducer_fastpath_sub = $self->{_current_reducer_fastpath_sub};
        };
        # pass the data to be reduced
	&{$reducer_fastpath_sub}($fref);
    };

    croak $self->{_prog} . ": no input data at all.\n"
        if (!defined($last_key));
    # print STDERR "done with input, last_key=$last_key\n";
    # close out any pending processing? (use the force option)
    $self->_close_old_key($last_key, 1);

    # reap sorter, if necessary
    # xxx: no longer, because this hangs (sometimes)!
    # see explanation in $self->setup.
#     if (!$self->{_pre_sorted}) {
#	$self->{_in_thread}->join;
#    };
}

=head2 finish

    $filter->finish();

Internal: write trailer.

=cut

sub finish ($) {
    my($self) = @_;

    # Let the postprocessor cleanup things.
    # It holds the output handle and calls the REAL finish.
    #
    # We just join on them to make sure we don't terminate the 
    # main thread prematurely.
    $self->{_reducer_postprocess_thread}->join;
}


=head1 AUTHOR and COPYRIGHT

Copyright (C) 1991-2008 by John Heidemann <johnh@isi.edu>

This program is distributed under terms of the GNU general
public license, version 2.  See the file COPYING
with the distribution for details.

=cut

1;
